#ifndef __FFLASFFPACK_WINOTHRESHOLD
#define __FFLASFFPACK_WINOTHRESHOLD 1000
#endif

#ifndef __FFLASFFPACK_WINOTHRESHOLD_FLT
#define __FFLASFFPACK_WINOTHRESHOLD_FLT 2000
#endif

#ifndef __FFLASFFPACK_WINOTHRESHOLD_BAL
#define __FFLASFFPACK_WINOTHRESHOLD_BAL 1000
#endif

#ifndef __FFLASFFPACK_WINOTHRESHOLD_BAL_FLT
#define __FFLASFFPACK_WINOTHRESHOLD_BAL_FLT 2000
#endif

#ifndef __FFLASFFPACK_PLUQ_THRESHOLD
#define __FFLASFFPACK_PLUQ_THRESHOLD 256
#endif

#ifndef __FFLASFFPACK_LUDIVINE_THRESHOLD
#define __FFLASFFPACK_LUDIVINE_THRESHOLD 1
#endif

#ifndef __FFLASFFPACK_CHARPOLY_LUKrylov_ArithProg_THRESHOLD
#define __FFLASFFPACK_CHARPOLY_LUKrylov_ArithProg_THRESHOLD 1000
#endif

#ifndef __FFLASFFPACK_CHARPOLY_Danilevskii_LUKrylov_THRESHOLD
#define __FFLASFFPACK_CHARPOLY_Danilevskii_LUKrylov_THRESHOLD 16
#endif

#ifndef __FFLASFFPACK_ARITHPROG_THRESHOLD
#define __FFLASFFPACK_ARITHPROG_THRESHOLD 30
#endif

#ifndef __FFLASFFPACK_FTRTRI_THRESHOLD
#define __FFLASFFPACK_FTRTRI_THRESHOLD 32
#endif

#ifndef __FFLASFFPACK_FSYTRF_THRESHOLD
#define __FFLASFFPACK_FSYTRF_THRESHOLD 64
#endif

#ifndef __FFLASFFPACK_FSYRK_THRESHOLD
#define __FFLASFFPACK_FSYRK_THRESHOLD 3000
#endif
/* -*- mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
// vim:sts=4:sw=4:ts=4:et:sr:cino=>s,f0,{0,g0,(0,\:0,t0,+0,=s
